@@ -10,6 +10,7 @@
 import pytest
 
 from sphinx import addnodes
+from sphinx.addnodes import desc
 from sphinx.domains.c import DefinitionParser, DefinitionError
 from sphinx.domains.c import _max_id, _id_prefix, Symbol
 from sphinx.testing import restructuredtext
@@ -27,17 +28,29 @@ def parse(name, string):
     return ast
 
 
-def check(name, input, idDict, output=None):
+def _check(name, input, idDict, output, key, asTextOutput):
+    if key is None:
+        key = name
+    key += ' '
+    if name in ('function', 'member'):
+        inputActual = input
+        outputAst = output
+        outputAsText = output
+    else:
+        inputActual = input.format(key='')
+        outputAst = output.format(key='')
+        outputAsText = output.format(key=key)
+    if asTextOutput is not None:
+        outputAsText = asTextOutput
+
     # first a simple check of the AST
-    if output is None:
-        output = input
-    ast = parse(name, input)
+    ast = parse(name, inputActual)
     res = str(ast)
-    if res != output:
+    if res != outputAst:
         print("")
         print("Input:    ", input)
         print("Result:   ", res)
-        print("Expected: ", output)
+        print("Expected: ", outputAst)
         raise DefinitionError("")
     rootSymbol = Symbol(None, None, None, None)
     symbol = rootSymbol.add_declaration(ast, docname="TestDoc")
@@ -45,6 +58,13 @@ def check(name, input, idDict, output=None):
     signode = addnodes.desc_signature(input, '')
     parentNode += signode
     ast.describe_signature(signode, 'lastIsName', symbol, options={})
+    resAsText = parentNode.astext()
+    if resAsText != outputAsText:
+        print("")
+        print("Input:    ", input)
+        print("astext(): ", resAsText)
+        print("Expected: ", outputAsText)
+        raise DefinitionError("")
 
     idExpected = [None]
     for i in range(1, _max_id + 1):
@@ -77,6 +97,17 @@ def check(name, input, idDict, output=None):
         raise DefinitionError("")
 
 
+def check(name, input, idDict, output=None, key=None, asTextOutput=None):
+    if output is None:
+        output = input
+    # First, check without semicolon
+    _check(name, input, idDict, output, key, asTextOutput)
+    if name != 'macro':
+        # Second, check with semicolon
+        _check(name, input + ' ;', idDict, output + ';', key,
+           asTextOutput + ';' if asTextOutput is not None else None)
+
+
 def test_expressions():
     def exprCheck(expr, output=None):
         class Config:
@@ -96,6 +127,15 @@ def test_expressions():
             print("Result:   ", res)
             print("Expected: ", output)
             raise DefinitionError("")
+        displayString = ast.get_display_string()
+        if res != displayString:
+            # note: if the expression contains an anon name then this will trigger a falsely
+            print("")
+            print("Input:    ", expr)
+            print("Result:   ", res)
+            print("Display:  ", displayString)
+            raise DefinitionError("")
+
     # type expressions
     exprCheck('int*')
     exprCheck('int *const*')
@@ -226,24 +266,24 @@ def test_expressions():
 
 
 def test_type_definitions():
-    check('type', "T", {1: "T"})
+    check('type', "{key}T", {1: "T"})
 
-    check('type', "bool *b", {1: 'b'})
-    check('type', "bool *const b", {1: 'b'})
-    check('type', "bool *const *b", {1: 'b'})
-    check('type', "bool *volatile *b", {1: 'b'})
-    check('type', "bool *restrict *b", {1: 'b'})
-    check('type', "bool *volatile const b", {1: 'b'})
-    check('type', "bool *volatile const b", {1: 'b'})
-    check('type', "bool *volatile const *b", {1: 'b'})
-    check('type', "bool b[]", {1: 'b'})
-    check('type', "long long int foo", {1: 'foo'})
+    check('type', "{key}bool *b", {1: 'b'}, key='typedef')
+    check('type', "{key}bool *const b", {1: 'b'}, key='typedef')
+    check('type', "{key}bool *const *b", {1: 'b'}, key='typedef')
+    check('type', "{key}bool *volatile *b", {1: 'b'}, key='typedef')
+    check('type', "{key}bool *restrict *b", {1: 'b'}, key='typedef')
+    check('type', "{key}bool *volatile const b", {1: 'b'}, key='typedef')
+    check('type', "{key}bool *volatile const b", {1: 'b'}, key='typedef')
+    check('type', "{key}bool *volatile const *b", {1: 'b'}, key='typedef')
+    check('type', "{key}bool b[]", {1: 'b'}, key='typedef')
+    check('type', "{key}long long int foo", {1: 'foo'}, key='typedef')
     # test decl specs on right
-    check('type', "bool const b", {1: 'b'})
+    check('type', "{key}bool const b", {1: 'b'}, key='typedef')
 
     # from breathe#267 (named function parameters for function pointers
-    check('type', 'void (*gpio_callback_t)(struct device *port, uint32_t pin)',
-          {1: 'gpio_callback_t'})
+    check('type', '{key}void (*gpio_callback_t)(struct device *port, uint32_t pin)',
+          {1: 'gpio_callback_t'}, key='typedef')
 
 
 def test_macro_definitions():
@@ -256,6 +296,10 @@ def test_macro_definitions():
     check('macro', 'M(arg, ...)', {1: 'M'})
     check('macro', 'M(arg1, arg2, ...)', {1: 'M'})
     check('macro', 'M(arg1, arg2, arg3, ...)', {1: 'M'})
+    # GNU extension
+    check('macro', 'M(arg1, arg2, arg3...)', {1: 'M'})
+    with pytest.raises(DefinitionError):
+        check('macro', 'M(arg1, arg2..., arg3)', {1: 'M'})
 
 
 def test_member_definitions():
@@ -370,28 +414,34 @@ def test_function_definitions():
           output='void f(int arr[static volatile const 42])')
 
 
-def test_union_definitions():
-    check('struct', 'A', {1: 'A'})
+def test_nested_name():
+    check('struct', '{key}.A', {1: "A"})
+    check('struct', '{key}.A.B', {1: "A.B"})
+    check('function', 'void f(.A a)', {1: "f"})
+    check('function', 'void f(.A.B a)', {1: "f"})
 
 
 def test_union_definitions():
-    check('union', 'A', {1: 'A'})
+    check('struct', '{key}A', {1: 'A'})
+
+
+def test_union_definitions():
+    check('union', '{key}A', {1: 'A'})
 
 
 def test_enum_definitions():
-    check('enum', 'A', {1: 'A'})
+    check('enum', '{key}A', {1: 'A'})
 
-    check('enumerator', 'A', {1: 'A'})
-    check('enumerator', 'A = 42', {1: 'A'})
+    check('enumerator', '{key}A', {1: 'A'})
+    check('enumerator', '{key}A = 42', {1: 'A'})
 
 
 def test_anon_definitions():
-    return  # TODO
-    check('class', '@a', {3: "Ut1_a"})
-    check('union', '@a', {3: "Ut1_a"})
-    check('enum', '@a', {3: "Ut1_a"})
-    check('class', '@1', {3: "Ut1_1"})
-    check('class', '@a::A', {3: "NUt1_a1AE"})
+    check('struct', '@a', {1: "@a"}, asTextOutput='struct [anonymous]')
+    check('union', '@a', {1: "@a"}, asTextOutput='union [anonymous]')
+    check('enum', '@a', {1: "@a"}, asTextOutput='enum [anonymous]')
+    check('struct', '@1', {1: "@1"}, asTextOutput='struct [anonymous]')
+    check('struct', '@a.A', {1: "@a.A"}, asTextOutput='struct [anonymous].A')
 
 
 def test_initializers():
@@ -424,6 +474,8 @@ def test_attributes():
     check('member', '__attribute__(()) int f', {1: 'f'})
     check('member', '__attribute__((a)) int f', {1: 'f'})
     check('member', '__attribute__((a, b)) int f', {1: 'f'})
+    check('member', '__attribute__((optimize(3))) int f', {1: 'f'})
+    check('member', '__attribute__((format(printf, 1, 2))) int f', {1: 'f'})
     # style: user-defined id
     check('member', 'id_attr int f', {1: 'f'})
     # style: user-defined paren
@@ -485,6 +537,16 @@ def test_build_domain_c(app, status, warning):
     assert len(ws) == 0
 
 
+@pytest.mark.sphinx(testroot='domain-c', confoverrides={'nitpicky': True})
+def test_build_domain_c_namespace(app, status, warning):
+    app.builder.build_all()
+    ws = filter_warnings(warning, "namespace")
+    assert len(ws) == 0
+    t = (app.outdir / "namespace.html").read_text()
+    for id_ in ('NS.NSVar', 'NULLVar', 'ZeroVar', 'NS2.NS3.NS2NS3Var', 'PopVar'):
+        assert 'id="c.{}"'.format(id_) in t
+
+
 @pytest.mark.sphinx(testroot='domain-c', confoverrides={'nitpicky': True})
 def test_build_domain_c_anon_dup_decl(app, status, warning):
     app.builder.build_all()
@@ -494,6 +556,13 @@ def test_build_domain_c_anon_dup_decl(app, status, warning):
     assert "WARNING: c:identifier reference target not found: @b" in ws[1]
 
 
+@pytest.mark.sphinx(testroot='domain-c', confoverrides={'nitpicky': True})
+def test_build_domain_c_semicolon(app, status, warning):
+    app.builder.build_all()
+    ws = filter_warnings(warning, "semicolon")
+    assert len(ws) == 0
+
+
 def test_cfunction(app):
     text = (".. c:function:: PyObject* "
             "PyType_GenericAlloc(PyTypeObject *type, Py_ssize_t nitems)")
@@ -526,3 +595,13 @@ def test_cvar(app):
     domain = app.env.get_domain('c')
     entry = domain.objects.get('PyClass_Type')
     assert entry == ('index', 'c.PyClass_Type', 'var')
+
+
+def test_noindexentry(app):
+    text = (".. c:function:: void f()\n"
+            ".. c:function:: void g()\n"
+            "   :noindexentry:\n")
+    doctree = restructuredtext.parse(app, text)
+    assert_node(doctree, (addnodes.index, desc, addnodes.index, desc))
+    assert_node(doctree[0], addnodes.index, entries=[('single', 'f (C function)', 'c.f', '', None)])
+    assert_node(doctree[2], addnodes.index, entries=[])
